/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/DungeonModel.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

DungeonModel * DungeonModel_create() {
	stemobject_create_implementation(DungeonModel, init)
}

void DungeonModel_init(compat_type(DungeonModel *) selfPtr) {
	DungeonModel * self = selfPtr;
	
	StemObject_init(self);
	self->dispose = DungeonModel_dispose;
	
	self->startRoom = NULL;
	self->startPosition.x = 0.0f;
	self->startPosition.y = 0.0f;
	self->startFacing = SOUTH;
	self->roomCount = 0;
	self->roomFileNames = NULL;
	self->exitCount = 0;
	self->exits = NULL;
}

void DungeonModel_dispose(compat_type(DungeonModel *) selfPtr) {
	DungeonModel * self = selfPtr;
	size_t roomIndex;
	
	for (roomIndex = 0; roomIndex < self->roomCount; roomIndex++) {
		free(self->roomFileNames[roomIndex]);
	}
	free(self->roomFileNames);
	free(self->exits);
	StemObject_dispose(self);
}

DungeonModel * DungeonModel_deserialize(compat_type(DeserializationContext *) deserializationContext) {
	stemobject_deserialize_implementation(DungeonModel, deserializationContext, loadSerializedData)
}

bool DungeonModel_loadSerializedData(compat_type(DungeonModel *) selfPtr, compat_type(DeserializationContext *) deserializationContext) {
	DungeonModel * self = selfPtr;
	DeserializationContext * context = deserializationContext;
	uint16_t formatVersion;
	Vector2 startPosition;
	Direction startFacing;
	size_t roomCount, roomIndex;
	char ** roomFileNames;
	size_t exitCount, exitIndex;
	struct DungeonModel_exit * exits;
	const char * startRoom, * fileName, * exitName;
	
	context->beginStructure(context, "dungeon");
	formatVersion = context->readUInt16(context, "format_version");
	if (context->status == SERIALIZATION_ERROR_OK && formatVersion != DUNGEON_SERIALIZATION_FORMAT_VERSION) {
		return false;
	}
	
	context->beginStructure(context, "start");
	startRoom = context->readString(context, "room");
	context->beginStructure(context, "position");
	startPosition.x = context->readFloat(context, "x");
	startPosition.y = context->readFloat(context, "y");
	context->endStructure(context);
	startFacing = context->readEnumeration(context, "facing", enumKV(NORTH), enumKV(NORTHEAST), enumKV(EAST), enumKV(SOUTHEAST), enumKV(SOUTH), enumKV(SOUTHWEST), enumKV(WEST), enumKV(NORTHWEST), NULL);
	context->endStructure(context);
	roomCount = context->beginArray(context, "rooms");
	roomFileNames = calloc(sizeof(char *), roomCount);
	for (roomIndex = 0; roomIndex < roomCount; roomIndex++) {
		fileName = context->readString(context, NULL);
		roomFileNames[roomIndex] = malloc(strlen(fileName) + 1);
		strcpy(roomFileNames[roomIndex], fileName);
	}
	context->endArray(context);
	exitCount = context->beginDictionary(context, "exits");
	exits = calloc(sizeof(struct DungeonModel_exit), exitCount);
	for (exitIndex = 0; exitIndex < exitCount; exitIndex++) {
		exitName = context->readNextDictionaryKey(context);
		exits[exitIndex].name = Atom_fromString(exitName);
		context->beginStructure(context, exitName);
		exits[exitIndex].room = Atom_fromString(context->readString(context, "room"));
		if (!strcmp(exits[exitIndex].room, "__DUNGEON_COMPLETE__")) {
			exits[exitIndex].position.x = exits[exitIndex].position.y = 0.0f;
		} else {
			context->beginStructure(context, "position");
			exits[exitIndex].position.x = context->readFloat(context, "x");
			exits[exitIndex].position.y = context->readFloat(context, "y");
			context->endStructure(context);
		}
		context->endStructure(context);
	}
	context->endDictionary(context);
	context->endStructure(context);
	
	if (context->status != SERIALIZATION_ERROR_OK) {
		for (roomIndex = 0; roomIndex < roomCount; roomIndex++) {
			free(roomFileNames[roomIndex]);
		}
		free(roomFileNames);
		free(exits);
		return false;
	}
	
	DungeonModel_init(self);
	self->startRoom = Atom_fromString(startRoom);
	self->startPosition = startPosition;
	self->startFacing = startFacing;
	self->roomCount = roomCount;
	self->roomFileNames = roomFileNames;
	self->exitCount = exitCount;
	self->exits = exits;
	
	return true;
}

void DungeonModel_serialize(compat_type(DungeonModel *) selfPtr, compat_type(SerializationContext *) serializationContext) {
	DungeonModel * self = selfPtr;
	SerializationContext * context = serializationContext;
	size_t roomIndex, exitIndex;
	
	context->beginStructure(context, "dungeon");
	context->writeUInt16(context, "format_version", DUNGEON_SERIALIZATION_FORMAT_VERSION);
	context->beginStructure(context, "start");
	context->writeString(context, "room", self->startRoom);
	context->beginStructure(context, "position");
	context->writeFloat(context, "x", self->startPosition.x);
	context->writeFloat(context, "y", self->startPosition.y);
	context->endStructure(context);
	context->writeEnumeration(context, "facing", self->startFacing, enumKV(NORTH), enumKV(NORTHEAST), enumKV(EAST), enumKV(SOUTHEAST), enumKV(SOUTH), enumKV(SOUTHWEST), enumKV(WEST), enumKV(NORTHWEST), NULL);
	context->endStructure(context);
	context->beginArray(context, "rooms");
	for (roomIndex = 0; roomIndex < self->roomCount; roomIndex++) {
		context->writeString(context, NULL, self->roomFileNames[roomIndex]);
	}
	context->endArray(context);
	context->beginDictionary(context, "exits");
	for (exitIndex = 0; exitIndex < self->exitCount; exitIndex++) {
		context->beginStructure(context, self->exits[exitIndex].name);
		context->writeString(context, "room", self->exits[exitIndex].room);
		if (strcmp(self->exits[exitIndex].room, "__DUNGEON_COMPLETE__")) {
			context->beginStructure(context, "position");
			context->writeFloat(context, "x", self->exits[exitIndex].position.x);
			context->writeFloat(context, "y", self->exits[exitIndex].position.y);
			context->endStructure(context);
		}
		context->endStructure(context);
	}
	context->endDictionary(context);
	context->endStructure(context);
}
