/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/DungeonDoorModel.h"
#include "convergence/DungeonKeyModel.h"
#include "convergence/EnemyBerserkerModel.h"
#include "convergence/EnemyFirstBossModel.h"
#include "convergence/EnemyRangerModel.h"
#include "convergence/GameStateModel.h"
#include "convergence/HealPowerupModel.h"
#include "convergence/PlayerBerserkerModel.h"
#include "convergence/PlayerRangerModel.h"
#include "convergence/PlayerSpyModel.h"
#include "convergence/PlayerDuelistModel.h"
#include "convergence/PlayerSpeedsterModel.h"
#include "convergence/RankUpPowerupModel.h"
#include "convergence/RoomExitModel.h"
#include "convergence/StaircaseModel.h"
#include <stdio.h>
#include <string.h>

GameStateModel * GameStateModel_create(DungeonModel * dungeon, enum PlayerClass playerClass, ResourceManager * resourceManager) {
	stemobject_create_implementation(GameStateModel, init, dungeon, playerClass, resourceManager)
}

static bool playerDeath(void * sender, Atom eventID, void * eventData, void * context) {
	GameStateModel * self = context;
	
	self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(GAME_STATE_EVENT_PLAYER_DIED), NULL);
	return true;
}

static bool playerExited(void * sender, Atom eventID, void * eventData, void * context) {
	GameStateModel * self = context;
	
	self->exitName = eventData;
	return false;
}

void GameStateModel_init(compat_type(GameStateModel *) selfPtr, DungeonModel * dungeon, enum PlayerClass playerClass, ResourceManager * resourceManager) {
	GameStateModel * self = selfPtr;
	DungeonRoomModel * room;
	DungeonRoomStateModel * roomState;
	size_t roomIndex, enemyIndex, objectIndex, exitIndex;
	
	StemObject_init(self);
	self->dispose = GameStateModel_dispose;
	self->exit = GameStateModel_exit;
	self->restartRoom = GameStateModel_restartRoom;
	
	self->eventDispatcher = EventDispatcher_create(self);
	self->dungeon = dungeon;
	self->roomCount = dungeon->roomCount;
	self->rooms = malloc(sizeof(DungeonRoomStateModel *) * self->roomCount);
	self->activeRoom = NULL;
	self->exitName = NULL;
	for (roomIndex = 0; roomIndex < self->roomCount; roomIndex++) {
		room = resourceManager->referenceResource(resourceManager, "dungeonRoom", dungeon->roomFileNames[roomIndex]);
		roomState = DungeonRoomStateModel_create(room->name, room->tileCountX, room->tileCountY, room->tiles, room->tileCanvasFileName);
		for (enemyIndex = 0; enemyIndex < room->enemyCount; enemyIndex++) {
			compat_type(EnemyModel *) enemy;
			
			switch (room->enemies[enemyIndex].type) {
				case DUNGEON_ENEMY_TYPE_BERSERKER:
					enemy = EnemyBerserkerModel_create(roomState, room->enemies[enemyIndex].rank, room->enemies[enemyIndex].startPosition, room->enemies[enemyIndex].startFacing, room->enemies[enemyIndex].inventoryItemCount, room->enemies[enemyIndex].inventoryItems, room->enemies[enemyIndex].patrolWaypointCount, room->enemies[enemyIndex].patrolWaypoints);
					break;
				case DUNGEON_ENEMY_TYPE_RANGER:
					enemy = EnemyRangerModel_create(roomState, room->enemies[enemyIndex].rank, room->enemies[enemyIndex].startPosition, room->enemies[enemyIndex].startFacing, room->enemies[enemyIndex].inventoryItemCount, room->enemies[enemyIndex].inventoryItems, room->enemies[enemyIndex].patrolWaypointCount, room->enemies[enemyIndex].patrolWaypoints);
					break;
				case DUNGEON_ENEMY_TYPE_FIRST_BOSS:
					enemy = EnemyFirstBossModel_create(roomState, room->enemies[enemyIndex].rank, room->enemies[enemyIndex].startPosition, room->enemies[enemyIndex].startFacing, room->enemies[enemyIndex].inventoryItemCount, room->enemies[enemyIndex].inventoryItems, room->enemies[enemyIndex].patrolWaypointCount, room->enemies[enemyIndex].patrolWaypoints);
					break;
				default:
					continue;
			}
			roomState->addObject(roomState, enemy);
		}
		for (objectIndex = 0; objectIndex < room->objectCount; objectIndex++) {
			switch (room->objects[objectIndex].type) {
				case DUNGEON_OBJECT_TYPE_HEAL_POWERUP:
					roomState->addObject(roomState, HealPowerupModel_create(roomState, room->objects[objectIndex].u.healPowerup.rank, room->objects[objectIndex].u.healPowerup.position));
					break;
				case DUNGEON_OBJECT_TYPE_DOOR:
					roomState->addObject(roomState, DungeonDoorModel_create(roomState, room->objects[objectIndex].u.door.keyID, room->objects[objectIndex].u.door.bounds));
					break;
				case DUNGEON_OBJECT_TYPE_KEY:
					roomState->addObject(roomState, DungeonKeyModel_create(roomState, room->objects[objectIndex].u.key.keyID, room->objects[objectIndex].u.key.position));
					break;
				case DUNGEON_OBJECT_TYPE_RANK_UP_POWERUP:
					roomState->addObject(roomState, RankUpPowerupModel_create(roomState, room->objects[objectIndex].u.rankUpPowerup.position));
					break;
				case DUNGEON_OBJECT_TYPE_STAIRCASE:
					roomState->addObject(roomState, StaircaseModel_create(roomState, room->objects[objectIndex].u.staircase.position, room->objects[objectIndex].u.staircase.down));
					break;
			}
		}
		for (exitIndex = 0; exitIndex < room->exitCount; exitIndex++) {
			roomState->addObject(roomState, RoomExitModel_create(roomState, room->exits[exitIndex].name, room->exits[exitIndex].bounds));
		}
		if (room->name == dungeon->startRoom) {
			compat_type(PlayerModel *) playerModel;
			
			self->activeRoomIndex = roomIndex;
			self->activeRoom = roomState;
			self->roomCopy = self->activeRoom->copy(self->activeRoom);
			switch (playerClass) {
				case PLAYER_CLASS_BERSERKER:
				default:
					playerModel = PlayerBerserkerModel_create(roomState, dungeon->startPosition, dungeon->startFacing);
					break;
				case PLAYER_CLASS_RANGER:
					playerModel = PlayerRangerModel_create(roomState, dungeon->startPosition, dungeon->startFacing);
					break;
				case PLAYER_CLASS_SPY:
					playerModel = PlayerSpyModel_create(roomState, dungeon->startPosition, dungeon->startFacing);
					break;
				case PLAYER_CLASS_DUELIST:
					playerModel = PlayerDuelistModel_create(roomState, dungeon->startPosition, dungeon->startFacing);
					break;
				case PLAYER_CLASS_SPEEDSTER:
					playerModel = PlayerSpeedsterModel_create(roomState, dungeon->startPosition, dungeon->startFacing);
					break;
			}
			self->playerModel = playerModel;
			self->playerModelCopy = self->playerModel->copy(self->playerModel);
			self->playerModel->eventDispatcher->registerForEvent(self->playerModel->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), playerDeath, self);
			self->playerModel->eventDispatcher->registerForEvent(self->playerModel->eventDispatcher, Atom_fromString(PLAYER_EVENT_EXITED), playerExited, self);
			roomState->addObject(roomState, self->playerModel);
		}
		self->rooms[roomIndex] = roomState;
		resourceManager->releaseResource(resourceManager, "dungeonRoom", dungeon->roomFileNames[roomIndex]);
	}
	if (self->activeRoom == NULL) {
		fprintf(stderr, "Error: Dungeon specifies starting room ID \"%s\", which doesn't exist\n", dungeon->startRoom);
	}
}

void GameStateModel_dispose(compat_type(GameStateModel *) selfPtr) {
	GameStateModel * self = selfPtr;
	size_t roomIndex;
	
	self->eventDispatcher->dispose(self->eventDispatcher);
	for (roomIndex = 0; roomIndex < self->roomCount; roomIndex++) {
		self->rooms[roomIndex]->dispose(self->rooms[roomIndex]);
	}
	free(self->rooms);
	self->roomCopy->dispose(self->roomCopy);
	self->playerModelCopy->dispose(self->playerModelCopy);
	StemObject_dispose(self);
}

void GameStateModel_exit(compat_type(GameStateModel *) selfPtr) {
	GameStateModel * self = selfPtr;
	size_t exitIndex;
	
	for (exitIndex = 0; exitIndex < self->dungeon->exitCount; exitIndex++) {
		if (self->dungeon->exits[exitIndex].name == self->exitName) {
			size_t roomIndex;
			
			if (!strcmp(self->dungeon->exits[exitIndex].room, "__DUNGEON_COMPLETE__")) {
				self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(GAME_STATE_EVENT_DUNGEON_COMPLETE), NULL);
				
			} else {
				for (roomIndex = 0; roomIndex < self->roomCount; roomIndex++) {
					if (self->rooms[roomIndex]->name == self->dungeon->exits[exitIndex].room) {
						self->activeRoom->removeObject(self->activeRoom, self->playerModel, false);
						self->activeRoom->softReset(self->activeRoom);
						self->activeRoomIndex = roomIndex;
						self->activeRoom = self->rooms[roomIndex];
						self->playerModel->position = self->playerModel->lastPosition = self->dungeon->exits[exitIndex].position;
						self->playerModel->stateModel = self->activeRoom;
						
						self->roomCopy->dispose(self->roomCopy);
						self->roomCopy = self->activeRoom->copy(self->activeRoom);
						self->playerModelCopy->dispose(self->playerModelCopy);
						self->playerModelCopy = self->playerModel->copy(self->playerModel);
						
						self->activeRoom->addObject(self->activeRoom, self->playerModel);
						self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(GAME_STATE_EVENT_ROOM_CHANGED), NULL);
						break;
					}
				}
				if (roomIndex == self->roomCount) {
					fprintf(stderr, "Warning: Unable to find room \"%s\" for exit \"%s\"\n", self->dungeon->exits[exitIndex].room, self->exitName);
				}
			}
			break;
		}
	}
	if (exitIndex == self->dungeon->exitCount) {
		fprintf(stderr, "Warning: Unable to perform exit \"%s\" in room \"%s\"\n", self->exitName, self->activeRoom->name);
	}
	self->exitName = NULL;
}

// HACK: This probably won't work if there are any sword swings or sparkle effects active in the room, but fortunately there shouldn't ever be when it's called
void GameStateModel_restartRoom(compat_type(GameStateModel *) selfPtr) {
	GameStateModel * self = selfPtr;
	DungeonRoomStateModel * oldRoom;
	
	oldRoom = self->rooms[self->activeRoomIndex];
	self->rooms[self->activeRoomIndex] = self->roomCopy->copy(self->roomCopy);
	self->activeRoom = self->rooms[self->activeRoomIndex];
	self->playerModel = self->playerModelCopy->copy(self->playerModelCopy);
	self->playerModel->stopMoving(self->playerModel);
	self->playerModel->stateModel = self->activeRoom;
	self->playerModel->eventDispatcher->registerForEvent(self->playerModel->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), playerDeath, self);
	self->playerModel->eventDispatcher->registerForEvent(self->playerModel->eventDispatcher, Atom_fromString(PLAYER_EVENT_EXITED), playerExited, self);
	self->activeRoom->addObject(self->activeRoom, self->playerModel);
	self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(GAME_STATE_EVENT_ROOM_CHANGED), NULL);
	oldRoom->dispose(oldRoom);
}
