/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/AudioManager.h"
#include "convergence/GameplayCamera.h"
#include "convergence/GameplayState.h"
#include "convergence/GraphicsConstants.h"
#include "convergence/PlaybackPlayerInputController.h"
#include "convergence/RecordingPlayerInputController.h"
#include "convergence/SharedEvents.h"
#include "convergence/ShellStateGlobals.h"
#include "shell/Shell.h"
#include "shell/ShellKeyCodes.h"

#ifdef STEM_PLATFORM_macosx
#include "nsopenglshell/NSOpenGLTarget.h"
#endif

#define DEATH_FADE_DURATION 1.5
#define RESPAWN_FADE_DURATION 0.25
#define DUNGEON_COMPLETE_FADE_IN_DURATION 1.25
#define DUNGEON_COMPLETE_FADE_OUT_DURATION 0.75

GameplayState * GameplayState_create(GameSession * gameSession) {
	stemobject_create_implementation(GameplayState, init, gameSession)
}

static void run(void * context) {
	GameplayState * self = context;
	
	self->playerInputController->update(self->playerInputController);
	self->gameStateController->update(self->gameStateController);
}

void GameplayState_init(compat_type(GameplayState *) selfPtr, GameSession * gameSession) {
	GameplayState * self = selfPtr;
	
	State_init(self);
	
	self->dispose = GameplayState_dispose;
	self->activate = GameplayState_activate;
	self->deactivate = GameplayState_deactivate;
	self->draw = GameplayState_draw;
	
	self->gameSession = gameSession;
	self->runLoop = FixedIntervalRunLoop_create(Shell_getCurrentTime, 1.0f / 120.0f, run, self);
	self->runLoop->pause(self->runLoop);
	self->gameStateModel = NULL;
	self->gameStateController = NULL;
	self->playerInputController = NULL;
	self->backgrounded = false;
}

void GameplayState_dispose(compat_type(GameplayState *) selfPtr) {
	GameplayState * self = selfPtr;
	
	self->runLoop->dispose(self->runLoop);
	self->view->dispose(self->view);
	
	State_dispose(selfPtr);
}

static void pause(GameplayState * self) {
	self->paused = true;
	self->pauseItemIndex = 0;
	self->runLoop->pause(self->runLoop);
	AudioManager_pauseMusic();
	
#ifdef STEM_PLATFORM_macosx
	NSOpenGLShell_setCursorVisible(true);
#endif
}

static void unpause(GameplayState * self) {
	self->paused = false;
	self->lastDrawTime = Shell_getCurrentTime();
	self->runLoop->resume(self->runLoop);
	AudioManager_resumeMusic();
	Shell_redisplay();
	
#ifdef STEM_PLATFORM_macosx
	if (Shell_isFullScreen()) {
		NSOpenGLShell_setCursorVisible(false);
	}
#endif
}

static bool mouseDown(void * sender, Atom eventID, void * eventData, void * context) {
	GameplayState * self = context;
	struct mouseEvent * event = eventData;
	float localY;
	
	if (event->button == 0 && self->paused) {
		localY = (g_windowHeight - event->y) / g_windowHeight * 2.0f - 1.0f;
		if (localY <= 0.0625f && localY > -0.0625f) {
			unpause(self);
		} else if (localY <= -0.0625f && localY > -0.1875f) {
			self->stateController->transition(self->stateController, "game_over");
		}
	}
	return true;
}

static bool keyDown(void * sender, Atom eventID, void * eventData, void * context) {
	struct keyEvent * event = eventData;
	GameplayState * self = context;
	bool handled = false;
	
	if (self->paused) {
		switch (event->keyCode) {
			case KEYBOARD_ESCAPE:
				unpause(self);
				handled = true;
				break;
				
			case KEYBOARD_UP_ARROW:
			case KEYBOARD_DOWN_ARROW:
				self->pauseItemIndex = (self->pauseItemIndex + 1) % 2;
				Shell_redisplay();
				handled = true;
				break;
				
			case KEYBOARD_RETURN_OR_ENTER:
			case KEYBOARD_SPACEBAR:
				switch (self->pauseItemIndex) {
					case 0:
						unpause(self);
						break;
						
					case 1:
						self->stateController->transition(self->stateController, "game_over");
						break;
				}
				handled = true;
				break;
		}
		
	} else {
		handled = self->playerInputController->keyDown(self->playerInputController, event->charCode, event->keyCode, event->modifiers);
		if (handled) {
			self->runLoop->resume(self->runLoop);
			Shell_redisplay();
		}
		
		if (event->keyCode == KEYBOARD_ESCAPE) {
			pause(self);
			handled = true;
		}
	}
	
	return handled;
}

static bool keyUp(void * sender, Atom eventID, void * eventData, void * context) {
	struct keyEvent * event = eventData;
	GameplayState * self = context;
	
	return self->playerInputController->keyUp(self->playerInputController, event->charCode, event->keyCode, event->modifiers);
}

static bool keyModifiersChanged(void * sender, Atom eventID, void * eventData, void * context) {
	struct keyEvent * event = eventData;
	GameplayState * self = context;
	
	return self->playerInputController->keyModifiersChanged(self->playerInputController, event->modifiers);
}

static bool resized(void * sender, Atom eventID, void * eventData, void * context) {
	Shell_redisplay();
	return true;
}

static bool backgrounded(void * sender, Atom eventID, void * eventData, void * context) {
	GameplayState * self = context;
	
	self->backgrounded = true;
	self->runLoop->pause(self->runLoop);
	AudioManager_pauseMusic();
	return true;
}

static bool foregrounded(void * sender, Atom eventID, void * eventData, void * context) {
	GameplayState * self = context;
	
	self->backgrounded = false;
	if (!self->paused) {
		self->lastDrawTime = Shell_getCurrentTime();
		self->runLoop->resume(self->runLoop);
		AudioManager_resumeMusic();
		Shell_redisplay();
	}
	return true;
}

static bool fullscreened(void * sender, Atom eventID, void * eventData, void * context) {
#ifdef STEM_PLATFORM_macosx
	GameplayState * self = context;
	
	if (Shell_isFullScreen() && !self->paused) {
		NSOpenGLShell_setCursorVisible(false);
	}
#endif
	return true;
}

static bool windowed(void * sender, Atom eventID, void * eventData, void * context) {
#ifdef STEM_PLATFORM_macosx
	NSOpenGLShell_setCursorVisible(true);
#endif
	return true;
}

static bool playerDied(void * sender, Atom eventID, void * eventData, void * context) {
	GameplayState * self = context;
	
	AudioManager_stopMusic();
	self->playerDied = true;
	self->playerDeathTime = 0.0;
	self->view->playFadeEffect(self->view, 0x00000000, 0x000000FF, DEATH_FADE_DURATION);
	return true;
}

static bool roomChanged(void * sender, Atom eventID, void * eventData, void * context) {
	GameplayState * self = context;
	
	// Sort of hack; will be resumed on next redisplay
	self->runLoop->pause(self->runLoop);
	Shell_redisplay();
	
	self->view->dispose(self->view);
	self->view = GameplayStateView_create(self->gameStateModel, GameplayCamera_create(self->gameStateModel->playerModel), self->gameSession->resourceManager);
	return true;
}

static bool dungeonComplete(void * sender, Atom eventID, void * eventData, void * context) {
	GameplayState * self = context;
	
	AudioManager_play("level_complete");
	AudioManager_stopMusic();
	self->dungeonComplete = true;
	self->dungeonFadeInComplete = false;
	self->dungeonCompleteTime = 0.0;
	self->view->playFadeEffect(self->view, 0xFFFFFF00, 0xFFFFFFFF, DUNGEON_COMPLETE_FADE_IN_DURATION);
	return true;
}

void GameplayState_activate(compat_type(GameplayState *) selfPtr) {
	GameplayState * self = selfPtr;
	DungeonModel * dungeon;
	
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_MOUSE_DOWN), mouseDown, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_DOWN), keyDown, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_UP), keyUp, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_MODIFIERS_CHANGED), keyModifiersChanged, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_RESIZED), resized, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_BACKGROUNDED), backgrounded, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_FOREGROUNDED), foregrounded, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_FULLSCREENED), fullscreened, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_WINDOWED), windowed, self);
	
	self->playerDied = false;
	self->dungeonComplete = false;
	self->paused = false;
	dungeon = self->gameSession->resourceManager->referenceResource(self->gameSession->resourceManager, "dungeon", "dungeons/dungeon0/dungeon0.json");
	self->gameStateModel = GameStateModel_create(dungeon, self->gameSession->playerClass, self->gameSession->resourceManager);
	self->gameStateModel->eventDispatcher->registerForEvent(self->gameStateModel->eventDispatcher, Atom_fromString(GAME_STATE_EVENT_PLAYER_DIED), playerDied, self);
	self->gameStateModel->eventDispatcher->registerForEvent(self->gameStateModel->eventDispatcher, Atom_fromString(GAME_STATE_EVENT_ROOM_CHANGED), roomChanged, self);
	self->gameStateModel->eventDispatcher->registerForEvent(self->gameStateModel->eventDispatcher, Atom_fromString(GAME_STATE_EVENT_DUNGEON_COMPLETE), dungeonComplete, self);
	self->gameStateController = GameStateController_create(self->gameStateModel);
	
	if (self->gameSession->recordInputToFile != NULL) {
		self->playerInputController = (PlayerInputController *) RecordingPlayerInputController_create(self->gameStateModel->playerModel, self->gameSession->playerInputMap);
	} else if (self->gameSession->playInputFromFile != NULL) {
		self->playerInputController = (PlayerInputController *) PlaybackPlayerInputController_create(self->gameStateModel->playerModel, self->gameSession->playerInputMap, self->gameSession->playInputFromFile);
	} else {
		self->playerInputController = PlayerInputController_create(self->gameStateModel->playerModel, self->gameSession->playerInputMap);
	}
	self->lastDrawTime = Shell_getCurrentTime();
	self->view = GameplayStateView_create(self->gameStateModel, GameplayCamera_create(self->gameStateModel->playerModel), self->gameSession->resourceManager);
	self->pauseScreenView = PauseScreenView_create(self->gameSession->resourceManager, &self->pauseItemIndex);
	
	self->runLoop->resume(self->runLoop);
	AudioManager_resumeMusic();
	AudioManager_startMusic("music/dungeon.ogg");
	Shell_redisplay();
	
#ifdef STEM_PLATFORM_macosx
	if (Shell_isFullScreen()) {
		NSOpenGLShell_setCursorVisible(false);
	}
#endif
}

void GameplayState_deactivate(compat_type(GameplayState *) selfPtr) {
	GameplayState * self = selfPtr;
	
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_MOUSE_DOWN), mouseDown, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_DOWN), keyDown, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_UP), keyUp, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_MODIFIERS_CHANGED), keyModifiersChanged, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_RESIZED), resized, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_BACKGROUNDED), backgrounded, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_FOREGROUNDED), foregrounded, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_FULLSCREENED), fullscreened, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_WINDOWED), windowed, self);
	
	self->pauseScreenView->dispose(self->pauseScreenView);
	self->view->dispose(self->view);
	if (self->gameSession->recordInputToFile != NULL) {
		((RecordingPlayerInputController *) self->playerInputController)->save(self->playerInputController, self->gameSession->recordInputToFile);
	}
	self->playerInputController->dispose(self->playerInputController);
	self->gameStateController->dispose(self->gameStateController);
	self->gameStateModel->dispose(self->gameStateModel);
	self->gameSession->resourceManager->releaseResource(self->gameSession->resourceManager, "dungeon", "dungeons/dungeon0/dungeon0.json");
	
	self->runLoop->pause(self->runLoop);
	AudioManager_stopMusic();
	
#ifdef STEM_PLATFORM_macosx
	NSOpenGLShell_setCursorVisible(true);
#endif
}

void GameplayState_draw(compat_type(GameplayState *) selfPtr) {
	GameplayState * self = selfPtr;
	double currentTime, deltaTime;
	
	currentTime = Shell_getCurrentTime();
	deltaTime = (self->paused || self->backgrounded) ? 0.0 : currentTime - self->lastDrawTime;
	self->lastDrawTime = currentTime;
	
	if (self->playerDied) {
		self->playerDeathTime += deltaTime;
		
		if (self->playerDeathTime >= DEATH_FADE_DURATION) {
			self->gameStateModel->restartRoom(self->gameStateModel);
			self->playerInputController->setPlayerModel(self->playerInputController, self->gameStateModel->playerModel);
			
			self->view->playFadeEffect(self->view, 0x000000FF, 0x00000000, RESPAWN_FADE_DURATION);
			self->playerDied = false;
			AudioManager_startMusic("music/dungeon.ogg");
		}
	}
	
	if (self->dungeonComplete) {
		self->dungeonCompleteTime += deltaTime;
		
		if (!self->dungeonFadeInComplete && self->dungeonCompleteTime >= DUNGEON_COMPLETE_FADE_IN_DURATION) {
			self->dungeonFadeInComplete = true;
			self->dungeonCompleteTime = 0.0;
			self->view->playFadeEffect(self->view, 0xFFFFFFFF, 0x000000FF, DUNGEON_COMPLETE_FADE_OUT_DURATION);
		}
		if (self->dungeonFadeInComplete && self->dungeonCompleteTime >= DUNGEON_COMPLETE_FADE_OUT_DURATION) {
			self->stateController->transition(self->stateController, "game_over");
			return;
		}
	}
	
	// HACK: Need to multithread!
	AudioManager_singleThreadStreamMusic();
	
	if (!self->dungeonComplete) {
		self->runLoop->run(self->runLoop);
	}
	
	self->view->draw(self->view, deltaTime);
	if (self->paused) {
		self->pauseScreenView->draw(self->pauseScreenView);
	}
	
	if (self->backgrounded || self->paused) {
		self->runLoop->pause(self->runLoop);
	} else {
		self->runLoop->resume(self->runLoop);
		Shell_redisplay();
	}
}
