/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/InventoryModel.h"
#include "convergence/SharedDefinitions.h"
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

InventoryModel * InventoryModel_create() {
	stemobject_create_implementation(InventoryModel, init)
}

InventoryModel * InventoryModel_createWithItems(size_t itemCount, struct InventoryModel_item * items) {
	stemobject_create_implementation(InventoryModel, initWithItems, itemCount, items)
}

void sharedInit(InventoryModel * self) {
	self->dispose = InventoryModel_dispose;
	self->copy = InventoryModel_copy;
	self->addItem = InventoryModel_addItem;
	self->hasKey = InventoryModel_hasKey;
	self->useKey = InventoryModel_useKey;
}

void InventoryModel_init(compat_type(InventoryModel *) selfPtr) {
	InventoryModel_initWithItems(selfPtr, 0, NULL);
}

void InventoryModel_initWithItems(compat_type(InventoryModel *) selfPtr, size_t itemCount, struct InventoryModel_item * items) {
	InventoryModel * self = selfPtr;
	
	StemObject_init(self);
	sharedInit(self);
	
	self->itemCount = itemCount;
	if (itemCount > 0) {
		self->items = malloc(sizeof(struct InventoryModel_item) * itemCount);
		memcpy(self->items, items, sizeof(struct InventoryModel_item) * itemCount);
	} else {
		self->items = NULL;
	}
}

void InventoryModel_initCopy(compat_type(InventoryModel *) selfPtr, compat_type(InventoryModel *) originalPtr) {
	InventoryModel * self = selfPtr, * original = originalPtr;
	
	StemObject_init(self);
	sharedInit(self);
	
	self->itemCount = original->itemCount;
	if (self->itemCount > 0) {
		self->items = malloc(sizeof(struct InventoryModel_item) * self->itemCount);
		memcpy(self->items, original->items, sizeof(struct InventoryModel_item) * self->itemCount);
	} else {
		self->items = NULL;
	}
}

void InventoryModel_dispose(compat_type(InventoryModel *) selfPtr) {
	InventoryModel * self = selfPtr;
	
	free(self->items);
	StemObject_dispose(self);
}

compat_type(InventoryModel *) InventoryModel_copy(compat_type(InventoryModel *) selfPtr) {
	stemobject_copy_implementation(InventoryModel, initCopy)
}

void InventoryModel_addItem(compat_type(InventoryModel *) selfPtr, InventoryItemType type, ...) {
	InventoryModel * self = selfPtr;
	struct InventoryModel_item item;
	va_list args;
	
	item.type = type;
	va_start(args, type);
	switch (type) {
		case INVENTORY_ITEM_TYPE_HEAL_POWERUP:
			item.u.healPowerup.rank = va_arg(args, int);
			break;
		case INVENTORY_ITEM_TYPE_KEY:
			item.u.key.keyID = va_arg(args, int);
			break;
		default:
			fprintf(stderr, "Warning: Unknown inventory item type %d\n", type);
			return;
	}
	va_end(args);
	self->items = realloc(self->items, sizeof(struct InventoryModel_item) * (self->itemCount + 1));
	self->items[self->itemCount++] = item;
}

bool InventoryModel_hasKey(compat_type(InventoryModel *) selfPtr, int keyID) {
	InventoryModel * self = selfPtr;
	size_t itemIndex;
	
	for (itemIndex = 0; itemIndex < self->itemCount; itemIndex++) {
		if (self->items[itemIndex].type == INVENTORY_ITEM_TYPE_KEY && self->items[itemIndex].u.key.keyID == keyID) {
			return true;
		}
	}
	return false;
}

void InventoryModel_useKey(compat_type(InventoryModel *) selfPtr, int keyID) {
	InventoryModel * self = selfPtr;
	size_t itemIndex;
	
	for (itemIndex = 0; itemIndex < self->itemCount; itemIndex++) {
		if (self->items[itemIndex].type == INVENTORY_ITEM_TYPE_KEY && self->items[itemIndex].u.key.keyID == keyID) {
			self->itemCount--;
			for (; itemIndex < self->itemCount; itemIndex++) {
				self->items[itemIndex] = self->items[itemIndex + 1];
			}
		}
	}
}
