/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/PlayerSpeedsterModel.h"
#include "convergence/ProjectileModel.h"
#include "convergence/SharedDefinitions.h"
#include "convergence/SwordSwingModel.h"

static const int DASH_COOLDOWNS[3] = {
	60,
	50,
	40
};
static const int DASH_DAMAGES[3] = {
	3,
	4,
	6
};
static const int DASH_DURATIONS[3] = {
	20,
	28,
	36
};
static const float DASH_SPEEDS[3] = {
	0.3125f,
	0.34375f,
	0.375f
};

PlayerSpeedsterModel * PlayerSpeedsterModel_create(DungeonRoomStateModel * stateModel, Vector2 position, Direction facing) {
	stemobject_create_implementation(PlayerSpeedsterModel, init, stateModel, position, facing)
}

static bool rankUp(void * sender, Atom eventID, void * eventData, void * context) {
	PlayerSpeedsterModel * self = context;
	
	switch (self->rank) {
		case 1:
			self->velocityMax = 17.0f / 256.0f;
			self->dashKnockback = 0.15625f;
			break;
		case 2:
			self->velocityMax = 18.0f / 256.0f;
			self->dashKnockback = 0.1875f;
			break;
	}
	return true;
}

static void sharedInit(PlayerSpeedsterModel * self) {
	self->dispose = PlayerSpeedsterModel_dispose;
	self->copy = PlayerSpeedsterModel_copy;
	self->update = PlayerSpeedsterModel_update;
	self->atRest = PlayerSpeedsterModel_atRest;
	self->protected_ivar(collisionInternal) = PlayerSpeedsterModel_collisionInternal;
	self->startAttack = PlayerSpeedsterModel_startAttack;
	self->protected_ivar(knockedBack) = PlayerSpeedsterModel_knockedBack;
	
	self->playerClass = PLAYER_CLASS_SPEEDSTER;
	self->acceleration = 4.0f / 256.0f;
	self->deceleration = 2.0f / 256.0f;
	self->velocityMax = 16.0f / 256.0f;
	self->dashKnockback = 0.125f;
	
	self->eventDispatcher->registerForEvent(self->eventDispatcher, Atom_fromString(PLAYER_EVENT_RANK_UP), rankUp, self);
}

void PlayerSpeedsterModel_init(compat_type(PlayerSpeedsterModel *) selfPtr, DungeonRoomStateModel * stateModel, Vector2 position, Direction facing) {
	PlayerSpeedsterModel * self = selfPtr;
	
	PlayerModel_init(self, stateModel, position, facing);
	sharedInit(self);
	self->attackCooldown = 0;
	self->dashDamage = 0;
}

void PlayerSpeedsterModel_initCopy(compat_type(PlayerSpeedsterModel *) selfPtr, compat_type(PlayerSpeedsterModel *) originalPtr) {
	PlayerSpeedsterModel * self = selfPtr, * original = originalPtr;
	
	PlayerModel_initCopy(self, original);
	sharedInit(self);
	self->attackCooldown = original->attackCooldown;
	self->dashDamage = original->dashDamage;
	self->dashKnockback = original->dashKnockback;
	self->velocityMax = original->velocityMax;
}

void PlayerSpeedsterModel_dispose(compat_type(PlayerSpeedsterModel *) selfPtr) {
	PlayerSpeedsterModel * self = selfPtr;
	
	PlayerModel_dispose(self);
}

compat_type(PlayerSpeedsterModel *) PlayerSpeedsterModel_copy(compat_type(PlayerSpeedsterModel *) selfPtr) {
	stemobject_copy_implementation(PlayerSpeedsterModel, initCopy)
}

void PlayerSpeedsterModel_update(compat_type(PlayerSpeedsterModel *) selfPtr) {
	PlayerSpeedsterModel * self = selfPtr;
	
	PlayerModel_update(self);
	
	if (self->attackCooldown > 0) {
		self->attackCooldown--;
		if (self->attackCooldown < DASH_COOLDOWNS[self->rank] - DASH_DURATIONS[self->rank]) {
			self->dashDamage = 0;
		}
	}
}

bool PlayerSpeedsterModel_atRest(compat_type(PlayerSpeedsterModel *) selfPtr) {
	PlayerSpeedsterModel * self = selfPtr;
	return self->attackCooldown == 0 &&
	       PlayerModel_atRest(self);
}

void PlayerSpeedsterModel_collisionInternal(compat_type(PlayerModel *) selfPtr, compat_type(GameObjectModel *) objectPtr, const struct collision collision, float time) {
	PlayerSpeedsterModel * self = selfPtr;
	GameObjectModel * object = objectPtr;
	
	PlayerModel_collisionInternal(self, objectPtr, collision, time);
	if (object != NULL && object->type == GAME_OBJECT_TYPE_ENEMY && self->dashDamage > 0) {
		float dot;
		
		dot = Vector2_dot(collision.normal, self->velocity);
		self->velocity.x -= collision.normal.x * (dot + self->dashKnockback);
		self->velocity.y -= collision.normal.y * (dot + self->dashKnockback);
	}
	if (object == NULL || object->type == GAME_OBJECT_TYPE_DOOR || object->type == GAME_OBJECT_TYPE_ENEMY || object->type == GAME_OBJECT_TYPE_ROOM_EXIT) {
		self->dashDamage = 0;
	}
}

void PlayerSpeedsterModel_startAttack(compat_type(PlayerSpeedsterModel *) selfPtr) {
	PlayerSpeedsterModel * self = selfPtr;
	
	if (self->attackCooldown == 0 && self->hitCooldown == 0) {
		Vector2 direction;
		
		self->attackCooldown = DASH_COOLDOWNS[self->rank];
		self->dashDamage = DASH_DAMAGES[self->rank];
		direction = directionToVector(self->facing);
		
		self->velocity.x = direction.x * DASH_SPEEDS[self->rank];
		self->velocity.y = direction.y * DASH_SPEEDS[self->rank];
		self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_EVENT_DASH), NULL);
	}
}

void PlayerSpeedsterModel_knockedBack(compat_type(PlayerSpeedsterModel *) selfPtr, Vector2 knockback) {
	PlayerSpeedsterModel * self = selfPtr;
	
	if (self->dashDamage > 0) {
		return;
	}
	PlayerModel_knockedBack(self, knockback);
}
